﻿SoDWhisper = LibStub("AceAddon-3.0"):NewAddon("SoDWhisper", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0", "LibWho-2.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SoDWhisper")
SoDWhisper.media = LibStub("LibSharedMedia-3.0", true)
SoDWhisper:SetDefaultModuleState(false)

local defaults = {
	profile = {
		modules = {},
		timeout = 360,
		classColor = true,
		whispersound = true,
		soundchannel = "Master",
		realIDName = "FIRST",
		linesShown = 8,
		BTNotify = true,
		paneColor = {0, 0, 0, 1.0},
		paneBackground = "Solid",
		useSound = true,
		Otherwhispersound = "SoDWhisper",
		Guildwhispersound = "SoDWhisper",
		Friendwhispersound = "SoDWhisper",
		paneBorder = "Blizzard Tooltip",
		paneBorderColor = {0.5, 0.5, 0.5, 0.8},
		paneFont = "Arial Narrow",
		fontSize = 15,
		timeFormat = "",
		timeColor = {.5,.5,.5,.5},
		colorIncoming = {1,0.5,1},
		colorOutgoing = {0.73,0.73,1},
		combatUpdate = true,
		colorBN = true,
		inputSticky = true,
		ignoreDBM = true,
	},
	factionrealm = {
		chatHistory = {},
	},
	global = {
		chatHistory = {},
	},
}

-- local
local _G = _G
local dbg
local dbr
local guildTimer = nil
local guildEventRosterUp = nil
local BNToonWho = 11
local OtherWho = 5
local connectedRealms
local SoDLocalClasses = {}
local SoDLocalCities = {}
local GetCHoutput = {}			-- function SoDWhisper:GetChatHistory
local BNnameNoBT = {}

-- variables
SoDWhisper.lastSender = nil
SoDWhisper.lastSenderPath = nil
SoDWhisper.sessionStart = 0
SoDWhisper.player = {}
SoDWhisper.createDisplayName = {}
SoDWhisper.LogFrame = {}
SoDWhisper.OnlineRegFriends = 0
SoDWhisper.OnlineGuildNum = 0
SoDWhisper.OnlineBNFriends = 0
SoDWhisper.timeframe = 1
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------- Initialisation Functions --------------------------------------------------------------------
function SoDWhisper:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("SoDWhisperDB", defaults, "Default")
	self.sessionStart = time() - 30
	self:SetupOptions()
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:OnEnable()
	-- Cache incoming and outgoing whispers
	self:RegisterEvent("CHAT_MSG_WHISPER", SoDWhisper.ChatEventIncoming)
	self:RegisterEvent("CHAT_MSG_WHISPER_INFORM", SoDWhisper.ChatEventOutgoing)
	self:RegisterEvent("CHAT_MSG_BN_WHISPER", SoDWhisper.BNChatEventIncoming)
	self:RegisterEvent("CHAT_MSG_BN_WHISPER_INFORM", SoDWhisper.BNChatEventOutgoing)
	
	self:RegisterBucketEvent({"PLAYER_GUILD_UPDATE"}, 4, SoDWhisper.GuildPlayerUpdate)
	self:RegisterBucketEvent({"FRIENDLIST_UPDATE"}, 4, SoDWhisper.UpdateRegFriends)
	self:RegisterBucketEvent({"BN_FRIEND_INFO_CHANGED", "BN_FRIEND_ACCOUNT_ONLINE", "BN_FRIEND_ACCOUNT_OFFLINE"}, 4, SoDWhisper.UpdateBNFriends)
	self:RegisterBucketEvent({"BN_DISCONNECTED"}, 4, SoDWhisper.UpdateBNDisconnected)
	self:RegisterBucketEvent({"BN_CONNECTED"}, 4, SoDWhisper.UpdateBNConnected)
	
	-- Schedules
	self:ScheduleRepeatingTimer("UpdateOtherToons", 30)
	
	-- Load Modules from whisp
	for i, v in self:IterateModules() do
		if self.db.profile.modules[i] ~= false then v:Enable() end
	end
	
	dbg = SoDWhisper.db.global
	dbr = SoDWhisper.db.factionrealm
	SoDWhisper.player = {class = _G.select(2,_G.UnitClass("player")), realm = _G.GetRealmName(), faction = _G.UnitFactionGroup("player"), name = _G.GetUnitName("player")}
	
	-- Get Connected Realm
	connectedRealms = _G.GetAutoCompleteRealms()
		
	-- Get and set-up localized class names
	local tempLocalizedClasses = {}
	_G.FillLocalizedClassList(tempLocalizedClasses, true)
	for i, v in pairs(tempLocalizedClasses) do
		SoDLocalClasses[v] = i
	end
	tempLocalizedClasses = {}
	_G.FillLocalizedClassList(tempLocalizedClasses, false)
	for i, v in pairs(tempLocalizedClasses) do
		SoDLocalClasses[v] = i
	end
	
	-- Get Localized Capital Cities
	SoDLocalCities[_G.GetMapNameByID(381)] = true	-- Darnassus
	SoDLocalCities[_G.GetMapNameByID(471)] = true	-- Exodar
	SoDLocalCities[_G.GetMapNameByID(341)] = true	-- Ironforge
	SoDLocalCities[_G.GetMapNameByID(321)] = true	-- Orgrimmar
	SoDLocalCities[_G.GetMapNameByID(301)] = true	-- Stormwind City
	SoDLocalCities[_G.GetMapNameByID(480)] = true	-- Silvermoon City
	SoDLocalCities[_G.GetMapNameByID(362)] = true	-- Thunder Bluff
	SoDLocalCities[_G.GetMapNameByID(382)] = true	-- Undercity
	SoDLocalCities[_G.GetMapNameByID(504)] = true	-- Dalaran
	SoDLocalCities[_G.GetMapNameByID(481)] = true	-- Shattrath City
	SoDLocalCities[_G.GetMapNameByID(903)] = true	-- Shrine of Two Moons
	SoDLocalCities[_G.GetMapNameByID(905)] = true	-- Shrine of Seven Stars
	
	-- Create Frames
	SoDWhisper.createDisplayName = SoDWhisper:ChangePlayerName()
	SoDWhisper.LogFrame = SoDWhisper:CreateLogFrame()
	
	-- Channel Sticky set-up
	if SoDWhisper.db.profile.inputSticky then
		ChatTypeInfo["WHISPER"].sticky = 0
		ChatTypeInfo["BN_WHISPER"].sticky = 0
	end
	
	-- Set-up on load
	SoDWhisper:UpdateOnLoad()
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:OnDisable()
	self:UnregisterAllEvents()
	-- UnLoad Modules  
	for k, v in self:IterateModules() do
		if self.db.profile.modules[k] ~= false then 
			v:Disable() 
		end
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:OnProfileChanged()
	for k, v in self:IterateModules() do
		if self.db.profile.modules[k] ~= false then 
			v:OnSkinUpdate() 
		end
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------- Events ---------------------------------------------------------------------------------
function SoDWhisper:ChatEventIncoming(msg, plr, _, _, _, _, _, _, _, _, _, GUID) 
	SoDWhisper:UpdateChatHistory(msg, plr, plr, true, nil, GUID)
end

function SoDWhisper:ChatEventOutgoing(msg, plr, _, _, _, _, _, _, _, _, _, GUID) 
	SoDWhisper:UpdateChatHistory(msg, plr, SoDWhisper.player.name, false, nil, GUID)
end

function SoDWhisper:BNChatEventIncoming(msg, plr, _, _, _, _, _, _, _, _, _, _, presID)
	SoDWhisper:UpdateChatHistory(msg, plr, plr, true, presID)
end

function SoDWhisper:BNChatEventOutgoing(msg, plr, _, _, _, _, _, _, _, _, _, _, presID)
	SoDWhisper:UpdateChatHistory(msg, plr, SoDWhisper.player.name, false, presID)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateOtherToons(setup)
	if not SoDWhisper.db.profile.combatUpdate and _G.UnitAffectingCombat("player") and not setup then return end
	for i,v in pairs(dbr.chatHistory) do
		if v.time[1] and not v.typeBNtoon and not v.typeGuild and not v.typeRegFriend then
			SoDWhisper:SetupOtherToonEntry(i, nil, nil, setup)
		end
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateRegFriends(setup)
	if not SoDWhisper.db.profile.combatUpdate and _G.UnitAffectingCombat("player") and not setup then return end
	for i,v in pairs(dbr.chatHistory) do
		if v.typeRegFriend then v.typeRegFriend = false end
	end
	
	_, SoDWhisper.OnlineRegFriends = _G.GetNumFriends()
    for i = 1, _G.GetNumFriends() do 
      SoDWhisper:SetupGuildOrFriendEntry(i, nil, setup)
    end
	
	if not setup then 
		SoDWhisper:SendMessage("SoDWhisper_MESSAGE") 
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GuildPlayerUpdate()
	if _G.IsInGuild() then 
		SoDWhisper:UpdateGuild()
	else
		for i, v in pairs(dbr.chatHistory) do
			if v.typeGuild then v.typeGuild = false end
		end
		if guildTimer then
			SoDWhisper:CancelTimer(guildTimer)
			guildTimer = nil
			SoDWhisper:UnregisterBucket(guildEventRosterUp)
			guildEventRosterUp = nil
		end
		SoDWhisper.OnlineGuildNum = 0
		SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateGuild(setup)
	if not SoDWhisper.db.profile.combatUpdate and _G.UnitAffectingCombat("player") and not setup then return end
	if _G.IsInGuild() then
		if not guildTimer then
			guildEventRosterUp = self:RegisterBucketEvent({"GUILD_ROSTER_UPDATE"}, 5, SoDWhisper.UpdateGuild)
			_G.GuildRoster()
			guildTimer = self:ScheduleRepeatingTimer(_G.GuildRoster, 35)
		end
		
		SoDWhisper.player.guild = _G.GetGuildInfo("player")
		for i, v in pairs(dbr.chatHistory) do
			if v.typeGuild then v.typeGuild = false end
		end
		
		_, _, SoDWhisper.OnlineGuildNum  = _G.GetNumGuildMembers()
		for i = 1, _G.GetNumGuildMembers() do
			SoDWhisper:SetupGuildOrFriendEntry(i, true)
		end
		
		if not setup then 
			SoDWhisper:SendMessage("SoDWhisper_MESSAGE") 
		end
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateBNDisconnected()
	for i,v in pairs(dbr.chatHistory) do
		if v.typeBNtoon or v.BT then 
			v.typeBNtoon = false 
			v.BT = nil
		end
	end
	for i,v in pairs(dbg.chatHistory) do
		if v.online then 
			v.online, v.status, v.zone, v.guild, v.faction, v.realm, v.toon, v.class, v.level = false, nil, nil, nil, nil, nil, nil, nil, nil
		end
	end
	SoDWhisper.OnlineBNFriends = 0
	SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateBNConnected()
	SoDWhisper.UpdateBNFriends(true)
	SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateBNFriends(setup)
	if not SoDWhisper.db.profile.combatUpdate and _G.UnitAffectingCombat("player") and not setup or not _G.BNFeaturesEnabledAndConnected() then return end
	for i,v in pairs(dbr.chatHistory) do
		if v.typeBNtoon or v.BT then 
			v.typeBNtoon = false 
			v.BT = nil
		end
	end
	if not setup then
		for i,v in pairs(dbg.chatHistory) do
			v.delCheck = false
		end
	end
	
	_, SoDWhisper.OnlineBNFriends = _G.BNGetNumFriends()
	for i = 1, _G.BNGetNumFriends() do
		SoDWhisper:SetupBNEntry(nil, nil, setup, i)
	end
	
	if not setup then
		for i,v in pairs(dbg.chatHistory) do
			if not v.delCheck and v.online then 
				v.online, v.status, v.zone, v.guild, v.faction, v.realm, v.toon, v.class, v.level = false, nil, nil, nil, nil, nil, nil, nil, nil
			end
		end
		SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
--------------------------------------------------------- Update and Set-up History ------------------------------------------------------------------------
function SoDWhisper:SetupOtherToonEntry(plr, whisper, GUID, setup)
	local classT	
	local answerT, toondisName = SoDWhisper:GetConnectedAnsOrName(plr)
	
	if GUID then
		classT = SoDWhisper:GetSetupClass(_G.select(2, _G.GetPlayerInfoByGUID(GUID)))
	else
		classT = SoDWhisper:GetSetupClass(_G.select(2, _G.UnitClass(plr)))
	end	
	
	if not dbr.chatHistory[plr] then
		dbr.chatHistory[plr] = {sender = {}, message = {}, time = {}, incoming = {}, tells = 0, name = toondisName, displayName = toondisName, class = classT, online = true, timeWho = time()}
		
		if answerT then 
			SoDWhisper:WhoUpdateServer(plr, true, OtherWho) 
		end
	elseif answerT then 
		if whisper then 
			dbr.chatHistory[plr].online = true 
		end
		if not dbr.chatHistory[plr].typeGuild then
			local ttime = time()
			local systime = dbr.chatHistory[plr].time[#dbr.chatHistory[plr].time]
			
			if ( (SoDWhisper.timeframe == 1 and systime < SoDWhisper.sessionStart) or (SoDWhisper.timeframe == 2 and ttime - systime > 3600) or (SoDWhisper.timeframe == 3 and ttime - systime > 86400) or (SoDWhisper.timeframe == 4 and ttime - systime > 604800) ) then
			else 
				SoDWhisper:WhoUpdateServer(plr, setup, OtherWho) 
			end
		end
	elseif classT then
		dbr.chatHistory[plr].online, dbr.chatHistory[plr].level = true, _G.UnitLevel(plr)
	else
		dbr.chatHistory[plr].online = false
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:SetupGuildOrFriendEntry(index, isGuild, setup)
	local plr, levelT, zoneName, isOnline, statusT, classT, isMobile
	
	if isGuild then
		plr, _, _, levelT, _, zoneName, _, _, isOnline, statusT, classT, _, _, isMobile = _G.GetGuildRosterInfo(index)
		if not plr then return end -- Some how one of these can be nil
		if not isOnline and not isMobile then
			if dbr.chatHistory[plr] then 
				dbr.chatHistory[plr].typeGuild = true
				if dbr.chatHistory[plr].online then
					dbr.chatHistory[plr].online, dbr.chatHistory[plr].status, dbr.chatHistory[plr].zone = false, nil, nil
				end
			end
			return
		end
	else
		plr, levelT, classT, zoneName, isOnline, statusT = _G.GetFriendInfo(index)
		if not plr then return end -- Some how one of these can be nil
		if not string.find(plr,"-") then 
			plr = plr.."-"..SoDWhisper.player.realm 
		end
		if not isOnline then
			if dbr.chatHistory[plr] then 
				dbr.chatHistory[plr].typeRegFriend = true
				if dbr.chatHistory[plr].online then 
					dbr.chatHistory[plr].online, dbr.chatHistory[plr].status, dbr.chatHistory[plr].zone = false, nil, nil
				end
			end
			return
		end
	end
	
	classT = SoDWhisper:GetSetupClass(classT)
	if statusT == 1 then 
		statusT = "<AFK>"
	elseif statusT == 2 then 
		statusT = "<DND>" 
	end
	if isMobile then 
		statusT = statusT and "<M>"..statusT or "<M>" 
	end
	
	if not dbr.chatHistory[plr] then
		local toondisName, _ = string.split("-", plr)
		dbr.chatHistory[plr] = {sender = {}, message = {}, time = {}, incoming = {}, tells = 0, name = toondisName, displayName = toondisName, timeWho = time()}
		setup = true
	end
	dbr.chatHistory[plr].online, dbr.chatHistory[plr].zone, dbr.chatHistory[plr].level, dbr.chatHistory[plr].status, dbr.chatHistory[plr].class = true, zoneName, levelT, statusT, classT
	if isGuild then 
		dbr.chatHistory[plr].typeGuild = true
		dbr.chatHistory[plr].guild = SoDWhisper.player.guild
	else 
		dbr.chatHistory[plr].typeRegFriend = true 
		SoDWhisper:WhoUpdateServer(plr, setup, BNToonWho) -- Get guild info
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:SetupBNEntry(BNETID, returnT, setup, index)
	local presID, plr, battleTag, isBattletagFriend, clientT, isOnline, isAFK, isDND, disName, noBattletag, statusT
	
	if BNETID then 
		presID, plr, battleTag, isBattletagFriend, _, _, clientT, isOnline, _, isAFK, isDND = _G.BNGetFriendInfoByID(BNETID)
	else 
		presID, plr, battleTag, isBattletagFriend, _, _, clientT, isOnline, _, isAFK, isDND = _G.BNGetFriendInfo(index) 
	end
	if not plr then return end
	
	if not isBattletagFriend then
		disName = SoDWhisper:GetRealIDShortN(plr)
		if not battleTag then 
			battleTag = plr 
			noBattletag = true 
		end
	else 
		if not battleTag then return end -- Some how this can be nil, very rare!!
		disName, _ = string.split("#", battleTag) 
	end
	
	-- Fix RealID names, Crash, and self on different player login
	if setup and dbg.chatHistory[battleTag] and dbg.chatHistory[battleTag].message[1] then
		if not dbg.chatHistory[battleTag].changed then 
			dbg.chatHistory[battleTag].displayName = disName 
		end
		for j=1, #dbg.chatHistory[battleTag].message do
			if dbg.chatHistory[battleTag].incoming[j] then 
				dbg.chatHistory[battleTag].sender[j] = dbg.chatHistory[battleTag].displayName
			else 
				dbg.chatHistory[battleTag].sender[j] = SoDWhisper.player.name 
			end
		end
	end
	
	if not isOnline then
		if dbg.chatHistory[battleTag] then 
			dbg.chatHistory[battleTag].delCheck = true
			if dbg.chatHistory[battleTag].online then 
				dbg.chatHistory[battleTag].online, dbg.chatHistory[battleTag].status, dbg.chatHistory[battleTag].zone, dbg.chatHistory[battleTag].guild, dbg.chatHistory[battleTag].faction, 
						dbg.chatHistory[battleTag].realm, dbg.chatHistory[battleTag].toon, dbg.chatHistory[battleTag].class, dbg.chatHistory[battleTag].level = 
						false, nil, nil, nil, nil, nil, nil, nil, nil
			end
		end
		return
	end
	
	if isAFK then 
		statusT = "<AFK>" 
	elseif isDND then 
		statusT = "<DND>" 
	end
	
	if not dbg.chatHistory[battleTag] then 
		dbg.chatHistory[battleTag] = {sender = {}, message = {}, time = {}, incoming = {}, tells = 0} 
	end
	local dbgH = dbg.chatHistory[battleTag]
	dbgH.name, dbgH.pID, dbgH.class, dbgH.client, dbgH.online, dbgH.status, dbgH.BT, dbgH.delCheck = disName, presID, classT, clientT, isOnline, statusT, battleTag, true
	if not dbgH.changed then 
		dbgH.displayName = disName 
	end
	
	local found
	for i = 1, _G.BNGetNumFriendToons(index or _G.BNGetFriendIndex(presID)) do
		local _, toonName, clientTT, realmName, _, factionT, _, classT, _, zoneName, levelT = _G.BNGetFriendToonInfo(index or _G.BNGetFriendIndex(presID), i)
		
		if toonName and clientTT == BNET_CLIENT_WOW then
			if realmName then 
				toonName = toonName.."-"..realmName 
			end
			classT = SoDWhisper:GetSetupClass(classT)
			
			if not found then 
				dbgH.toon, dbgH.client, dbgH.realm, dbgH.faction, dbgH.class, dbgH.zone, dbgH.level = toonName, clientTT, realmName, factionT, classT, zoneName, levelT 
			end
			
			-- Setup Character of BN account
			if factionT == SoDWhisper.player.faction then
				local answerT, toondisName = SoDWhisper:GetConnectedAnsOrName(toonName)
				if answerT or dbr.chatHistory[toonName] then 
					if not dbr.chatHistory[toonName] then
						dbr.chatHistory[toonName] = {sender = {}, message = {}, time = {}, incoming = {}, tells = 0, name = toondisName, displayName = toondisName, timeWho = time()}		
						setup = true
					end
					
					local dbrH = dbr.chatHistory[toonName]
					dbrH.BT, dbrH.online, dbrH.zone, dbrH.level, dbrH.typeBNtoon, dbrH.class = battleTag, isOnline, zoneName, levelT, true, classT
					if not dbrH.typeGuild then 
						dbrH.status = statusT 
					end
					if dbrH.message[1] and dbg.chatHistory[battleTag] then
						for j=1, #dbrH.message do
							if ( dbrH.incoming[j] ) then
								tinsert(dbgH.sender, dbgH.displayName)
							else
								tinsert(dbgH.sender, dbrH.sender[j])
							end
							tinsert(dbgH.message, dbrH.message[j])
							tinsert(dbgH.incoming, dbrH.incoming[j])
							tinsert(dbgH.time, dbrH.time[j])
						end
						dbgH.tells = dbgH.tells + dbrH.tells
						dbrH.sender, dbrH.message, dbrH.incoming, dbrH.time, dbrH.tells = {}, {}, {}, {}, 0
					end
					
					if answerT and not found and not dbr.chatHistory[toonName].typeGuild then 
						SoDWhisper:WhoUpdateServer(toonName, setup, BNToonWho) 
					elseif answerT and not found then 
						dbgH.guild = dbr.chatHistory[toonName].guild 
					end
				end
			end
			found = true
		end
	end
	if not found then 
		dbgH.toon, dbgH.realm, dbgH.faction, dbgH.class, dbgH.zone, dbgH.level = nil, nil, nil, nil, nil, nil 
	end
	if returnT then return battleTag, noBattletag end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------- Incoming/Outgoing----------------------------------------------------------------------------
function SoDWhisper:UpdateChatHistory(msg, plr, snd, incoming, BNETID, GUID)
	if not plr then return end
	if not BNETID and not string.find(plr,"-") then 
		plr = plr.."-"..SoDWhisper.player.realm 
	end
	local battleTag, noBattletag, add, path, name
	
	-- Check for spam
	if SoDWhisper:HideAddonMessage(msg) then return end
	
	-- Check if toon is a BN friend
	if not BNETID and dbr.chatHistory[plr] and dbr.chatHistory[plr].BT and _G.BNFeaturesEnabledAndConnected() then
		BNETID = SoDWhisper:CompareBattleTag(dbr.chatHistory[plr].BT)
		if not BNETID then
			dbr.chatHistory[plr].BT = nil
			dbr.typeBNtoon = false
		end
	end

	if BNETID then 
		battleTag, noBattletag = SoDWhisper:SetupBNEntry(BNETID, true)
		path, name = dbg, battleTag
	else 
		if not dbr.chatHistory[plr] then
			SoDWhisper:SetupOtherToonEntry(plr, true, GUID) 
		else
			add = true
		end
		path, name = dbr, plr
	end
	
	-- Set up tells for BN or for reg toon
	if snd == SoDWhisper.player.name then 
		path.chatHistory[name].tells = 0
	else 
		path.chatHistory[name].tells = path.chatHistory[name].tells + 1 
	end
	
	-- Sound
	if incoming and SoDWhisper.db.profile.useSound then 
		SoDWhisper:IMSound(path, name) 
	end
		
	-- LDB/Minimap assignment
	SoDWhisper.lastSender, SoDWhisper.lastSenderPath = name, path
	
	-- Insert message
	if incoming then 
		snd = path.chatHistory[name].displayName 
	end
	tinsert(path.chatHistory[name].sender, snd)
	tinsert(path.chatHistory[name].message, msg)
	tinsert(path.chatHistory[name].time, time())
	tinsert(path.chatHistory[name].incoming, incoming)
	
	-- Call after time insert so who updates correctly
	if add then 
		SoDWhisper:SetupOtherToonEntry(plr, true, GUID) 
	end
	
	-- Update modules
	SoDWhisper:SendMessage("SoDWhisper_MESSAGE") 
	SoDWhisper:SendMessage("SoDWhisper_UPDATE_EDITBOX")
	
	-- Report No battleTag
	if SoDWhisper.db.profile.BTNotify and noBattletag and not BNnameNoBT[battleTag] then
		BNnameNoBT[battleTag] = true
		_G.BNSendWhisper(BNETID, L["SoDWhisper: Please Set Up A BattleTag At Battle.net"])
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------ Who Query -------------------------------------------------------------------------------
function SoDWhisper:WhoUpdateServer(name, force, minutes)
	if not force then
		if (dbr.chatHistory[name].timeWho) > (time() - (minutes * 60)) then return end
	end
	 
	local disName, realmT = string.split("-", name) 
	if realmT ~= SoDWhisper.player.realm then 
		disName = name 
	end
	
	local wholib = SoDWhisper
	dbr.chatHistory[name].timeWho = time()
	wholib:UserInfo(disName, {queue = wholib.WHOLIB_QUEUE_QUIET, timeout = 0, flags = wholib.WHOLIB_FLAG_ALWAYS_CALLBACK, callback = 'ReturnedWhoData'})
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Get Who Data
function SoDWhisper:ReturnedWhoData(plrinfo)
	if not plrinfo then return end
	if not string.find(plrinfo.Name,"-") then 
		plrinfo.Name = plrinfo.Name.."-"..SoDWhisper.player.realm 
	end
	if not dbr.chatHistory[plrinfo.Name] then return end
	
	local dbrH = dbr.chatHistory[plrinfo.Name]
	if plrinfo.Online == true then
		local classT = SoDWhisper:GetSetupClass(plrinfo.Class)
		
		dbrH.class, dbrH.online, dbrH.zone, dbrH.level, dbrH.guild = classT, true, plrinfo.Zone, plrinfo.Level, plrinfo.Guild
		if plrinfo.Guild and dbrH.BT and dbg.chatHistory[dbrH.BT] then
			dbg.chatHistory[dbrH.BT].guild = plrinfo.Guild
		end
	elseif plrinfo.Online == false then
		dbrH.online, dbrH.zone = false, nil
	end
	
	-- LDB/Minimap Update
	SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
end

------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------- Tools ---------------------------------------------------------------------------------
-- Returns the combined recent messages from this player (little code from whisp)
function SoDWhisper:GetChatHistory(plr, max, path, newToOld)
	if not path.chatHistory[plr] then return "" end
	local msg = ""
	if path.chatHistory[plr] then
		local n = #path.chatHistory[plr].message
		
		-- Generate conversation
		for i= (n > max and n - max + 1) or 1, n  do
			local tempFTS = SoDWhisper:FormatTimeStamp(path.chatHistory[plr].time[i])
			local tempFPN = SoDWhisper:FormatPlayerName(path.chatHistory[plr].sender[i], plr, path)
			local tempHex = SoDWhisper:HexColor(path.chatHistory[plr].incoming[i] and SoDWhisper.db.profile.colorIncoming or SoDWhisper.db.profile.colorOutgoing)
			local tempColorMessage = SoDWhisper:Colorise(path.chatHistory[plr].message[i], tempHex)
			tinsert(GetCHoutput, tempFTS..tempFPN..": "..tempColorMessage)
		end
		
		-- Sort output
		for i = (newToOld and 1) or #GetCHoutput, (newToOld and #GetCHoutput) or 1, (newToOld and 1) or -1 do
			msg = msg..GetCHoutput[i]
			if (newToOld and i < #GetCHoutput) or (not newToOld and i > 1) then 
				msg = msg.."\n"
			end
		end
		_G.wipe(GetCHoutput)
	end
	return msg
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetRealIDShortN(plr)
	if SoDWhisper.db.profile.realIDName == "FIRST" then
		return gsub(plr, "|Kf", "|Kg")
	elseif SoDWhisper.db.profile.realIDName == "LAST" then
		return gsub(plr, "|Kf", "|Ks")
	else return plr end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetConnectedAnsOrName(name)
	local shortName, realm = string.split("-", name)
 	if connectedRealms then
		for i = 1, #connectedRealms do
 	 		if realm == connectedRealms[i] then 
				return true, shortName
			end
		end
	elseif SoDWhisper.player.realm == realm then 
		return true, shortName
	end
	
	if realm then 
		shortName = shortName.."-"..string.sub(realm, 1, 3) 
	end
    return false, shortName
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Compare BattleTag 
function SoDWhisper:CompareBattleTag(incBT)
	if not incBT then return nil, nil, nil end
	local presIDT, plrT, battletagT	
	
	for i = 1, BNGetNumFriends() do		
		presIDT, plrT, battletagT = _G.BNGetFriendInfo(i)
		
		if battletagT and battletagT == incBT then
			return presIDT, plrT, battletagT
		elseif plrT == incBT then
			return presIDT, plrT, battletagT
		end
	end
	return nil, nil, nil
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:HideAddonMessage(msg)
	if SoDWhisper.db.profile.ignoreDBM and (string.sub(msg, 1,18) == "<Deadly Boss Mods>" or string.sub(msg, 1,5) == "<DBM>") then
		return true
	end
	return false
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetSetupClass(class)
	if not class then return nil end
	if SoDLocalClasses[class] then
		class = SoDLocalClasses[class]
	end
	return class
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetTells(name, path)
	if not name or not path or not path.chatHistory[name] then return "" end
	if path.chatHistory[name].tells > 0 then return path.chatHistory[name].tells end
	return ""
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetStatusicon(name, path, size)
	if not name or not path or not path.chatHistory[name] or not path.chatHistory[name].status then return "" end
	local t
	if string.find(path.chatHistory[name].status, "<AFK>") then
		t = "|TInterface\\FriendsFrame\\StatusIcon-Away:"..size..":"..size.."|t"
	elseif string.find(path.chatHistory[name].status, "<DND>") then
		t = "|TInterface\\FriendsFrame\\StatusIcon-DnD:"..size..":"..size.."|t"
	end
	if string.find(path.chatHistory[name].status, "<M>") then
		t = t and "|TInterface\\ChatFrame\\UI-ChatIcon-ArmoryChat:"..size..":"..size.."|t"..t or "|TInterface\\ChatFrame\\UI-ChatIcon-ArmoryChat:"..size..":"..size.."|t"
	end
	return t or ""
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetDisplayNameForTooltip(name, path, isBN, needToon, clientOK, size)
	local plr = path.chatHistory[name]
	if isBN then
		if not needToon then
			if not clientOK then
				return (plr.online and SoDWhisper:GetStatusicon(name, path, size).."|c0000FAF6"..plr.displayName.."|r") or "|c00585858"..plr.displayName.."|r"
			else
				return (plr.online and SoDWhisper:GetStatusicon(name, path, size).."|c0000FAF6["..plr.displayName.."]|r") or "|c00585858["..plr.displayName.."]|r"
			end
		else
			if not plr.online then 
				return "|c00585858-|r" 
			end
			if plr.toon and plr.client and plr.client == BNET_CLIENT_WOW then
				local temp, _ = string.split("-", plr.toon)
				return SoDWhisper:FormatPlayerName(temp, name, path, true)
			elseif plr.client and clientOK then 
				return SoDWhisper:GetGameIcon(plr.client, size)
			else 
				return "-"
			end
		end
	else
		return (plr.online and SoDWhisper:GetStatusicon(name, path, size)..SoDWhisper:FormatPlayerName(plr.displayName, name, path)) or "|c00585858"..plr.displayName.."|r"
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetColoredLevel(level, online, BT, isEditBox)
	if not level then 
		return (isEditBox and "") or (online and "-") or "|c00585858-|r"
	end
	if not online then 
		if isEditBox then 
			return (BT and "") or "|c00585858"..level..": |r" 
		else
			return (BT and "|c00585858-|r") or "|c00585858"..level.."|r" 
		end
	end
	local color = _G.GetQuestDifficultyColor(level)
	if isEditBox then 
		return "|cff"..string.format("%02x%02x%02x", color.r*255, color.g*255, color.b*255)..level.."|r"..": "
	end
	return "|cff"..string.format("%02x%02x%02x", color.r*255, color.g*255, color.b*255)..level.."|r"
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetZoneColor(zone, online)
	if not zone or zone == "" or zone == " " then 
		return (online and "-") or "|c00585858-|r"
	end
	return (SoDLocalCities[zone] and "|c0000FF00"..zone.."|r") or "|c00FFFF00"..zone.."|r"
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:GetRealmForTooltip(name, clientOK, size, path)
	if not path.chatHistory[name].online then return "|c00585858-|r" end
	if not dbg.chatHistory[name] then
		local _, tempR = string.split("-", name)
		
		if SoDWhisper.player.faction == "Horde" then
			return "|c00FF0000"..tempR.."|r"
		elseif SoDWhisper.player.faction == "Alliance" then
			return "|c000099FF"..tempR.."|r"
		else 
			return tempR 
		end
	end
	
	local dbgH = dbg.chatHistory[name]
	if clientOK and dbgH.client ~= BNET_CLIENT_WOW then 
		return SoDWhisper:GetGameIcon(dbgH.client, size, true) 
	end
	if dbgH.realm then
		if dbgH.toon and dbgH.faction then
			if dbgH.faction == "Horde" then
				return "|c00FF0000"..dbgH.realm.."|r"
			elseif dbgH.faction == "Alliance" then
				return "|c000099FF"..dbgH.realm.."|r"
			else 
				return dbgH.realm 
			end
		else
			return dbgH.realm
		end
	end
	return "-"
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:FormatPlayerName(snd, plr, path, toon)
	if not snd then return end
	if not SoDWhisper.db.profile.colorBN and path == dbg and snd ~= SoDWhisper.player.name and not toon then
		return "|c0000FAF6["..snd.."]|r"
	end
	
	local class
	if SoDWhisper.player.name == snd then 
		class = SoDWhisper.player.class
	elseif path.chatHistory[plr] and path.chatHistory[plr].class then
		class = path.chatHistory[plr].class
	end
	if not SoDWhisper.db.profile.classColor or not class then 
		if path == dbg and snd ~= SoDWhisper.player.name and not toon then
			return "|c0000FAF6"..snd.."|r"
		end
		return "|cffffff00"..snd.."|r" 
	end
	
	local classColorTable = _G.RAID_CLASS_COLORS[class]
	if not classColorTable then
		if path == dbg and snd ~= SoDWhisper.player.name and not toon then
			return "|c0000FAF6"..snd.."|r"
		end
		return "|cffffff00"..snd.."|r"
	end
	return string.format("\124cff%.2x%.2x%.2x", classColorTable.r*255, classColorTable.g*255, classColorTable.b*255)..snd.."\124r" -- the format from whisp
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- A little code from Whisp
function SoDWhisper:FormatTimeStamp(t)
	if SoDWhisper.db.profile.timeFormat == "" then return "" end
	if t < time() - 86400 then
		t = date("%d/%m - "..SoDWhisper.db.profile.timeFormat, t)
	else
		t = date(SoDWhisper.db.profile.timeFormat, t)
	end
	return SoDWhisper:Colorise(t, SoDWhisper:HexColor(SoDWhisper.db.profile.timeColor)).." "
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Returns the hex value of the rgb
function SoDWhisper:HexColor(r, g, b)
	if type(r)=="table" then
		r, g, b = r[1], r[2], r[3]
	end
	return string.format("%02x%02x%02x", 255*r, 255*g, 255*b)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Color the text
function SoDWhisper:Colorise(text, hexcolor)
	text = gsub(text, "(|c.-|H.-|r)", "|r%1|cff"..hexcolor)
	return "|cff"..hexcolor..text.."|r"
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Sounds
function SoDWhisper:IMSound(path, name)
	if path == dbg or path.chatHistory[name].typeBNtoon or path.chatHistory[name].typeRegFriend then
		local sT = SoDWhisper.media:Fetch('sound', SoDWhisper.db.profile.Friendwhispersound)
		_G.PlaySoundFile(sT, SoDWhisper.db.profile.soundchannel) 
	elseif path.chatHistory[name].typeGuild then
		local sT = SoDWhisper.media:Fetch('sound', SoDWhisper.db.profile.Guildwhispersound)
		_G.PlaySoundFile(sT, SoDWhisper.db.profile.soundchannel)
	else
		local sT = SoDWhisper.media:Fetch('sound', SoDWhisper.db.profile.Otherwhispersound)
		_G.PlaySoundFile(sT, SoDWhisper.db.profile.soundchannel) 
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Get game icon and name
function SoDWhisper:GetGameIcon(client, size, noIcon)
	if not client then return "" end
	if( client == BNET_CLIENT_SC2 ) then
		return "|TInterface\\FriendsFrame\\Battlenet-Sc2icon:"..size..":"..size.."|t"
	elseif( client == BNET_CLIENT_D3 ) then
		return "|TInterface\\FriendsFrame\\Battlenet-D3icon:"..size..":"..size.."|t"
	elseif( client == BNET_CLIENT_WTCG ) then
		return "|TInterface\\FriendsFrame\\Battlenet-WTCGicon:"..size..":"..size.."|t"
	else
		return "|TInterface\\FriendsFrame\\Battlenet-Battleneticon:"..size..":"..size.."|t"
	end
end

------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------ FRAMES ----------------------------------------------------------------------------------
-- Frame ideas gathered from whisp
function SoDWhisper:ChangePlayerName()
	local frame = _G.CreateFrame("Frame", "SoDWhisper_USERINPUT", UIParent)
	frame:SetWidth(270)  
	frame:SetHeight(120)
	frame:SetPoint("CENTER", UIParent, "CENTER")
	frame:SetFrameLevel(_G.GameTooltip:GetFrameLevel() + 20)
	frame:SetFrameStrata("HIGH")
	frame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 3, right = 3, top = 3, bottom = 3},
	})
	frame:EnableMouse(true)
	frame:SetClampedToScreen(true)
	-- Set font
	frame.text = frame:CreateFontString("SoDWhisper_USERINPUTTEXT","ARTWORK","ChatFontNormal")
	frame.text:SetJustifyH("CENTER")
	frame.text:SetJustifyV("TOP")
	frame.text:SetTextColor(1.0,1.0,1.0,1.0)
	-- Set background-color
	frame:SetBackdropColor(0, 0, 0, 1)
	frame:SetBackdropBorderColor(0.5, 0.5, 0.5, 0.8)
	--------
	frame.titleRegion = frame:CreateTitleRegion()
	frame.titleRegion:SetPoint("BOTTOMLEFT", title, -5, -2)
	frame.titleRegion:SetPoint("TOPRIGHT", title, 5, 2)
	--------
	frame.inputBox = CreateFrame("EditBox", "SoDWhisper_USERINPUT", frame)
	frame.inputBox:SetWidth(250)
	frame.inputBox:SetHeight(30)
	frame.inputBox:SetPoint("CENTER", frame, "CENTER", 0, -10)
	frame.inputBox:SetFrameLevel(GameTooltip:GetFrameLevel() + 21)
	frame.inputBox:SetTextInsets(10, 10, 0, 0)
	frame.inputBox:SetAutoFocus(false)
	frame.inputBox:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 3, right = 3, top = 3, bottom = 3},
	})
	frame.inputBox:SetBackdropColor(0, 0, 0, 1)
	frame.inputBox:SetBackdropBorderColor(0.5, 0.5, 0.5, 0.8)
	frame.inputBox:SetScript("OnEscapePressed", function() 
		frame.inputBox:ClearFocus()
	end)
	
	--------
	frame.buttonEnter = CreateFrame("Button", nil, frame)
	frame.buttonEnter:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 10, 5)
	frame.buttonEnter:SetWidth(75)
	frame.buttonEnter:SetHeight(28)
	
	frame.buttonEnter:SetText("Enter")
	frame.buttonEnter:SetNormalFontObject("GameFontNormal")
	
	local ntex = frame.buttonEnter:CreateTexture()
	ntex:SetTexture("Interface/Buttons/UI-Panel-Button-Up")
	ntex:SetTexCoord(0, 0.625, 0, 0.6875)
	ntex:SetAllPoints()	
	frame.buttonEnter:SetNormalTexture(ntex)
	
	local htex = frame.buttonEnter:CreateTexture()
	htex:SetTexture("Interface/Buttons/UI-Panel-Button-Highlight")
	htex:SetTexCoord(0, 0.625, 0, 0.6875)
	htex:SetAllPoints()
	frame.buttonEnter:SetHighlightTexture(htex)
	-------
	frame.buttonReset = CreateFrame("Button", nil, frame)
	frame.buttonReset:SetPoint("BOTTOM", frame, "BOTTOM", 0, 5)
	frame.buttonReset:SetWidth(75)
	frame.buttonReset:SetHeight(28)
	
	frame.buttonReset:SetText("Reset")
	frame.buttonReset:SetNormalFontObject("GameFontNormal")
	
	ntex = frame.buttonReset:CreateTexture()
	ntex:SetTexture("Interface/Buttons/UI-Panel-Button-Up")
	ntex:SetTexCoord(0, 0.625, 0, 0.6875)
	ntex:SetAllPoints()	
	frame.buttonReset:SetNormalTexture(ntex)
	
	htex = frame.buttonReset:CreateTexture()
	htex:SetTexture("Interface/Buttons/UI-Panel-Button-Highlight")
	htex:SetTexCoord(0, 0.625, 0, 0.6875)
	htex:SetAllPoints()
	frame.buttonReset:SetHighlightTexture(htex)
	-------
	frame.buttonClose = CreateFrame("Button", nil, frame)
	frame.buttonClose:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -10, 5)
	frame.buttonClose:SetWidth(75)
	frame.buttonClose:SetHeight(28)
	
	frame.buttonClose:SetText("Close")
	frame.buttonClose:SetNormalFontObject("GameFontNormal")
	
	ntex = frame.buttonClose:CreateTexture()
	ntex:SetTexture("Interface/Buttons/UI-Panel-Button-Up")
	ntex:SetTexCoord(0, 0.625, 0, 0.6875)
	ntex:SetAllPoints()	
	frame.buttonClose:SetNormalTexture(ntex)
	
	htex = frame.buttonClose:CreateTexture()
	htex:SetTexture("Interface/Buttons/UI-Panel-Button-Highlight")
	htex:SetTexCoord(0, 0.625, 0, 0.6875)
	htex:SetAllPoints()
	frame.buttonClose:SetHighlightTexture(htex)
	SoDWhisper:SkinMyFrame(frame, nil, true)
	----------
	frame:Hide()
	return frame
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateChangePlayerName(frame, plr, path)
	frame:ClearAllPoints()
	frame:SetPoint("CENTER", UIParent, "CENTER")
	frame.text:SetText("|c0000FAF6SoDWhisper|r".."\n\n"..L["Display Name For:"].." "..path.chatHistory[plr].name)
	frame.text:SetPoint("TOP",frame,"TOP",0,-8)
	frame.buttonClose:SetButtonState("NORMAL", unlock)
	frame.buttonEnter:SetButtonState("NORMAL", unlock)
	frame.buttonReset:SetButtonState("NORMAL", unlock)
	frame:Show()
	frame.inputBox:SetFocus()
	frame.buttonClose:SetScript("OnMouseDown", function() 
		frame.inputBox:SetText("")
		frame:Hide()
	end)
	frame.buttonEnter:SetScript("OnMouseDown", function()
		local text = frame.inputBox:GetText()
		if string.len(text) < 13 and string.len(text) > 0 then
			path.chatHistory[plr].displayName = text
			for j=1, #path.chatHistory[plr].message do
				if path.chatHistory[plr].incoming[j] then
					path.chatHistory[plr].sender[j] = text
				end
			end
			
			path.chatHistory[plr].changed = true
			
			frame.inputBox:SetText("")
			frame:Hide()
			if SoDWhisper.lastSender == plr then
				SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
			end
		else print(L["Must be 1 - 12 characters in length"])
		end
	end)
	frame.buttonReset:SetScript("OnMouseDown", function() 
		frame.inputBox:SetText("")
		path.chatHistory[plr].displayName = path.chatHistory[plr].name
		for j=1, #path.chatHistory[plr].message do
			if path.chatHistory[plr].incoming[j] then
				path.chatHistory[plr].sender[j] = path.chatHistory[plr].name
			end
		end
		path.chatHistory[plr].changed = nil		
		frame:Hide()
		if SoDWhisper.lastSender == plr then
			SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
		end
	end)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:CreateMyFrame(name, width)
	local frame = _G.CreateFrame("Frame", "SoDWhisper_"..name, UIParent)
	frame:ClearAllPoints()
	-- Size is required to be set or SetPoint() won't work later on
	frame:SetWidth(width)  
	frame:SetHeight(20)
	frame:SetMinResize(200,20)
	frame:SetFrameLevel(_G.GameTooltip:GetFrameLevel() + 10)
	-- Setup textfields
	frame.text = frame:CreateFontString("SoDWhisper_"..name.."TEXT","ARTWORK","ChatFontNormal")
	frame.text:SetJustifyH("LEFT")
	frame.text:SetJustifyV("TOP")
	frame.text:SetTextColor(1.0,1.0,1.0,1.0)
	frame.text:SetNonSpaceWrap(true)
	--frame.text:SetIndentedWordWrap(true)
	-- Set background and border
	frame:SetFrameStrata("HIGH")
	SoDWhisper:SkinMyFrame(frame)
	frame:Hide()
	return frame
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:UpdateMyFrame(frame, plr, path, direction, ignore)
	frame.text:SetText(SoDWhisper:GetChatHistory(plr, SoDWhisper.db.profile.linesShown, path, direction))
	frame.text:SetPoint("TOPLEFT",frame,"TOPLEFT",7,-7)
	-- Adjust height of the frame to the height of the text
	frame.text:SetWidth(frame:GetRight() - frame:GetLeft() - 14)
	frame:SetHeight(frame.text:GetHeight()+16)
	-- Update tells on event
	if not ignore and path.chatHistory[plr].tells > 0 then
		path.chatHistory[plr].tells = 0
		SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:SkinMyFrame(frame, isLogFrame, isDisplayFrame)
	if not isLogFrame and not isDisplayFrame then
		local bg = SoDWhisper.media:Fetch('background', SoDWhisper.db.profile.paneBackground, "Blizzard Tooltip")
		local ed = SoDWhisper.media:Fetch('border', SoDWhisper.db.profile.paneBorder, "Blizzard Tooltip")
		frame:SetBackdrop({
			bgFile = bg, tile = true, tileSize = 16,
			edgeFile = ed, edgeSize = 16,
			insets = {left = 3, right = 3, top = 3, bottom = 3},
		})
		-- Set font
		frame.text:SetFont(SoDWhisper.media:Fetch('font', SoDWhisper.db.profile.paneFont), SoDWhisper.db.profile.fontSize)
		-- Set background-color
		local c,d = SoDWhisper.db.profile.paneColor, SoDWhisper.db.profile.paneBorderColor
		frame:SetBackdropColor(c[1], c[2], c[3], c[4])
		frame:SetBackdropBorderColor(d[1], d[2], d[3], d[4])
	end
	if isLogFrame then
		frame.text:SetFont(SoDWhisper.media:Fetch('font', SoDWhisper.db.profile.paneFont), 14)
		frame.logFrameInput:SetFont(SoDWhisper.media:Fetch('font', SoDWhisper.db.profile.paneFont), SoDWhisper.db.profile.fontSize)
	end
	if isDisplayFrame then
		frame.text:SetFont(SoDWhisper.media:Fetch('font', SoDWhisper.db.profile.paneFont), 14)
		frame.inputBox:SetFont(SoDWhisper.media:Fetch('font', SoDWhisper.db.profile.paneFont), 14)
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------- Log Window --------------------------------------------------------------------------------
function SoDWhisper:CreateLogFrame()
	local frame = _G.CreateFrame("Frame", "SoDWhisper_LogFrame1", UIParent)
	frame:SetWidth(700)  
	frame:SetHeight(510)
	frame:SetPoint("CENTER", UIParent, "CENTER")
	frame:SetFrameLevel(_G.GameTooltip:GetFrameLevel() + 30)
	frame:SetFrameStrata("HIGH")
	frame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 3, right = 3, top = 3, bottom = 3},
	})
	frame:EnableMouse(true)
	frame:SetClampedToScreen(true)
	-- Set font
	frame.text = frame:CreateFontString("SoDWhisper_USERINPUTTEXT","ARTWORK","ChatFontNormal")
	frame.text:SetJustifyH("CENTER")
	frame.text:SetJustifyV("TOP")
	frame.text:SetTextColor(1.0,1.0,1.0,1.0)
	frame.text:SetPoint("TOP",frame,"TOP",0,-15)
	-- Set background-color
	frame:SetBackdropColor(0, 0, 0, 1)
	frame:SetBackdropBorderColor(0.5, 0.5, 0.5, 0.8)
	--------
	frame.titleRegion = frame:CreateTitleRegion()
	frame.titleRegion:SetPoint("BOTTOMLEFT", title, -5, -2)
	frame.titleRegion:SetPoint("TOPRIGHT", title, 5, 2)
	--------
	frame.buttonClose = CreateFrame("Button", nil, frame)
	frame.buttonClose:SetPoint("BOTTOM", frame, "BOTTOM", 0, 10)
	frame.buttonClose:SetWidth(75)
	frame.buttonClose:SetHeight(28)
	
	frame.buttonClose:SetText("Close")
	frame.buttonClose:SetNormalFontObject("GameFontNormal")
	
	ntex = frame.buttonClose:CreateTexture()
	ntex:SetTexture("Interface/Buttons/UI-Panel-Button-Up")
	ntex:SetTexCoord(0, 0.625, 0, 0.6875)
	ntex:SetAllPoints()	
	frame.buttonClose:SetNormalTexture(ntex)
	
	htex = frame.buttonClose:CreateTexture()
	htex:SetTexture("Interface/Buttons/UI-Panel-Button-Highlight")
	htex:SetTexCoord(0, 0.625, 0, 0.6875)
	htex:SetAllPoints()
	frame.buttonClose:SetHighlightTexture(htex)
	--------
	frame.frame1 = _G.CreateFrame("Frame", "SoDWhisper_LogFrame2", frame)
	frame.frame1:SetWidth(625)  
	frame.frame1:SetHeight(425)
	frame.frame1:SetPoint("CENTER", frame, "CENTER")
	frame.frame1:SetFrameLevel(_G.GameTooltip:GetFrameLevel() + 31)
	frame.frame1:SetFrameStrata("HIGH")
	frame.frame1:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 3, right = 3, top = 3, bottom = 3},
	})
	frame.frame1:SetBackdropColor(0, 0, 0, 1)
	frame.frame1:SetBackdropBorderColor(0.5, 0.5, 0.5, 0.8)
	--------
	frame.logFrameInput = CreateFrame("EditBox", "SoDWhisper_LOGFRAMEINPUT", frame.frame1)
	frame.logFrameInput:SetMultiLine(true)
	frame.logFrameInput:SetAutoFocus(false)
	frame.logFrameInput:EnableMouse(true)
	--frame.logFrameInput:SetHyperlinksEnabled()
	--frame.logFrameInput:EnableKeyboard(false)
	frame.logFrameInput:SetWidth(620)
	frame.logFrameInput:SetHeight(425)
	frame.logFrameInput:SetFrameLevel(GameTooltip:GetFrameLevel() + 32)
	frame.logFrameInput:SetTextInsets(8, 28, 0, 0)
	frame.logFrameInput:SetScript("OnEscapePressed", function() 
		frame.logFrameInput:ClearFocus()
	end)
	SoDWhisper:SkinMyFrame(frame, true)
	--------
	frame.scroll = CreateFrame('ScrollFrame', 'nChatCopyScroll', frame.frame1, 'UIPanelScrollFrameTemplate')
	frame.scroll:SetPoint('TOPLEFT', frame.frame1, 'TOPLEFT', 0, -10)
	frame.scroll:SetPoint('BOTTOMRIGHT', frame.frame1, 'BOTTOMRIGHT', -30, 10)
	frame.scroll:SetScrollChild(frame.logFrameInput)
	
	frame:Hide()
	return frame
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function SoDWhisper:ShowLogFrame(frame, plr, path)
	if not plr or not path or not path.chatHistory[plr] or not path.chatHistory[plr].message[1] then return end
	frame:ClearAllPoints()
	frame:SetPoint("CENTER", UIParent, "CENTER")
	frame.buttonClose:SetButtonState("NORMAL", unlock)
	frame.text:SetText("|c0000FAF6SoDWhisper|r")
	frame.logFrameInput:SetText(SoDWhisper:GetChatHistory(plr, #path.chatHistory[plr].time, path, true))
	--frame.logFrameInput:SetFocus()
	--frame.scroll:SetVerticalScroll(frame.scroll:GetVerticalScrollRange())
	frame:Show()
	frame.buttonClose:SetScript("OnMouseDown", function() 
			frame.logFrameInput:SetText("")
			frame:Hide()
	end)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
--------------------------------------------------------------- OnLoad Function ----------------------------------------------------------------------------
function SoDWhisper:UpdateOnLoad() -- One Time on Load
	local t = SoDWhisper.sessionStart
	if SoDWhisper.db.profile.timeout > 0 then 
		t = time() - (3600 * SoDWhisper.db.profile.timeout)
	end

	for i,v in pairs(dbr.chatHistory) do
		if v.time[1] then
			local ct = v.time[#v.time]
			if ct and ct < t and not v.changed then
				dbr.chatHistory[i] = nil
			else
				for j=1, #v.time do
					if v.time[j] and v.time[j] < t then
						tremove(v.sender,j)
						tremove(v.message,j)
						tremove(v.time,j)
						tremove(v.incoming,j)
					end
				end
			end
		elseif not v.changed then 
			dbr.chatHistory[i] = nil 
		end
	end
	
	for i,v in pairs(dbg.chatHistory) do
		if v.time[1] then
			local ct = v.time[#v.time]
			if ct and ct < t and not v.changed then
				dbg.chatHistory[i] = nil
			else
				for j=1, #v.time do
					if v.time[j] and v.time[j] < t then
						tremove(v.sender,j)
						tremove(v.message,j)
						tremove(v.time,j)
						tremove(v.incoming,j)
					end
				end
			end
		elseif not v.changed then 
			dbg.chatHistory[i] = nil 
		end
	end
	
	-- Turn all offline
	for i,v in pairs(dbg.chatHistory) do
		v.online, v.zone, v.guild, v.faction, v.realm, v.toon, v.class, v.BT, v.delCheck = false, nil, nil, nil, nil, nil, nil, nil, false
	end
	for i,v in pairs(dbr.chatHistory) do
		v.online, v.zone, v.typeGuild, v.typeBNtoon, v.typeRegFriend = false, nil, nil, nil, nil
		if v.message[1] then
			for j=1, #v.message do
				if not v.incoming[j] then v.sender[j] = SoDWhisper.player.name end
			end
		end
	end
	
	SoDWhisper:UpdateGuild(true)
	SoDWhisper:UpdateBNFriends(true)	
	SoDWhisper:UpdateRegFriends(true)
	SoDWhisper:UpdateOtherToons(true)
	
	SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
end