local mod = SoDWhisper:NewModule("Editbox Plugin", "AceEvent-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SoDWhisper")
local QTC = LibStub('LibQTip-1.0')

mod.modName = L["Editbox Plugin"]

local _G = _G

local options = {
	editboxappearance = {
		type = "group",
		order = 5,
		name = L["Appearance"],
		disabled = function() return not mod:IsEnabled() end,
		args = {
			titlescale = {
				type = "range",
				order = 5,
				name = L["Title Scale"],
				get = function() return mod.db.profile.titlescale end,
				set = function(_,v) mod.db.profile.titlescale = v 
				end,
				min = .5, 
				max = 2, 
				step = .05,
			}, 	
			paneSortDown = {
				type = "toggle",
				order = 10,
				name = L["Message Sort"],
				desc = L["Show entries from old to new on the message panel for the EditBox."],
				get = function() return mod.db.profile.paneSortDown end,
				set = function(_,v) mod.db.profile.paneSortDown = v end,
			},
			combatHide = {
				type = "toggle",
				order = 15,
				name = L["Hide In Combat"],
				get = function() return mod.db.profile.hideInCombat end,
				set = function(_,v) mod.db.profile.hideInCombat = v end,
			},
		},
	},
	------------------------------------------------------------------------------------------------------------------------------------------------------------
	editboxposition = {
		type = "group",
		order = 10,
		name = L["Position"],
		disabled = function() return not mod:IsEnabled() end,
		args = {
			snap = {
				type = "toggle",
				order = 5,
				name = L["Snap To Inputbox"],
				desc = L["Makes the panel stick to the inputbox."],
				get = function() return mod.db.profile.snapToEditbox end,
				set = function(_,v) 
					mod.db.profile.snapToEditbox = v
					mod:SetPosition()
				end,
			},
			growup = {
				type = "toggle",
				order = 10,
				name = L["Grow Up"],
				desc = L["The panel will grow up from the set position."],
				get = function() return mod.db.profile.growUp end,
				set = function(_,v) 
					mod.db.profile.growUp = v 
					mod:SetPosition()
				end,
			}, 
			reset = {
				type = "execute",
				order = 15,
				name = L["Reset Position"],
				desc = L["Reset the position of the panel."],
				func = function() mod:ResetPosition() end,
			},
			lock = {
				type = "toggle",
				order = 20,
				name = L["Lock"],
				desc = L["Locks the panel."],
				get = function() return mod.db.profile.locked end,
				set = function(_,v) 
					mod.db.profile.locked = v
					if v == false then 
						mod.db.profile.snapToEditbox = false 
					end
				end,
			},
		},
	},
}

local defaults = {
	profile = {
		hideInCombat = false,              -- Hide the chatbox tooltip when in combat
		locked = true,                     -- Locks the chatbox pane in place
		frameWidth = 400,                  -- Set the width and position of the pane
		frameLeft = 10, 
		frameTop = 370, 
		frameBottom = 400,
		growUp = true,                     -- Grow the pane upwards
		snapToEditbox = true,              -- Snap the pane to the editbox
		titlescale = .9,
		paneSortDown = true,
	}
}
local editboxFrame = {}
local OEBName = nil
local OEBNamePath = nil
local inviteIcon = "|TInterface\\AddOns\\SoDWhisper\\MediaFiles\\inviteicon:17|t"
local addfriendIcon = "|TInterface\\AddOns\\SoDWhisper\\MediaFiles\\addfriendicon:17|t"
local ignoreIcon = "|TInterface\\AddOns\\SoDWhisper\\MediaFiles\\ignoreicon:17|t"
local locationIcon = "|TInterface\\AddOns\\SoDWhisper\\MediaFiles\\locationicon:17|t"
local copyIcon = "|TInterface\\AddOns\\SoDWhisper\\MediaFiles\\copyicon:10|t"
local dbg, dbr
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------- Initialisation Functions --------------------------------------------------------------------
function mod:OnInitialize()
	self.db = SoDWhisper.db:RegisterNamespace("Editbox Plugin", defaults)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:OnEnable()
	self:RegisterMessage("SoDWhisper_SKIN", "OnSkinUpdate")
	self:RegisterMessage("SoDWhisper_UPDATE_EDITBOX", "UpdateOpenEditBox")
	self:SecureHook("ChatEdit_UpdateHeader", "ChatEdit_Update")
	self:SecureHook("ChatEdit_DeactivateChat", "ChatEdit_Update")
	
	dbg, dbr = SoDWhisper.db.global, SoDWhisper.db.factionrealm
	
	editboxFrame = mod:CreateFrame()
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:OnDisable()
	mod:ChatEdit_OnHide()
	self:UnregisterAllEvents()
	self:UnhookAll()
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------- Edit-box Functions -----------------------------------------------------------------------
function mod:ChatEdit_Update(self)
	if not _G.ChatEdit_GetActiveWindow() or (mod.db.profile.hideInCombat and _G.InCombatLockdown()) then 
		mod:ChatEdit_OnHide()
		return 
	end
	
	local chatAttribute = self:GetAttribute("chatType")
	if chatAttribute == "WHISPER" then
	
		local tarAttribute = self:GetAttribute("tellTarget")
		if not string.find(tarAttribute,"-") then 
			tarAttribute = tarAttribute.."-"..SoDWhisper.player.realm 
		end
		
		if dbr.chatHistory[tarAttribute] then
			if dbr.chatHistory[tarAttribute].BT and _G.BNFeaturesEnabledAndConnected() then
				if dbg.chatHistory[dbr.chatHistory[tarAttribute].BT] and dbg.chatHistory[dbr.chatHistory[tarAttribute].BT].message[1] then
					if dbr.chatHistory[tarAttribute].BT == OEBName and dbg == OEBNamePath then return end  -- stops mod:ChatEdit_Update's rapid firing
					OEBName, OEBNamePath = dbr.chatHistory[tarAttribute].BT, dbg
					mod:ShowFrame()
				elseif dbr.chatHistory[tarAttribute].message[1] then
					if tarAttribute == OEBName and dbr == OEBNamePath then return end  -- stops mod:ChatEdit_Update's rapid firing
					OEBName, OEBNamePath = tarAttribute, dbr
					mod:ShowFrame()	
				else 
					mod:ChatEdit_OnHide() 
				end
			elseif dbr.chatHistory[tarAttribute].message[1] then
				if tarAttribute == OEBName and dbr == OEBNamePath then return end  -- stops mod:ChatEdit_Update's rapid firing
				OEBName, OEBNamePath = tarAttribute, dbr
				mod:ShowFrame()
			else 
				mod:ChatEdit_OnHide() 
			end
		else 
			mod:ChatEdit_OnHide() 
		end
	elseif chatAttribute == "BN_WHISPER" then
	
		local tarAttribute = self:GetAttribute("tellTarget")
		if _G.GetAutoCompletePresenceID(tarAttribute) then
			local _, plr, battletag = _G.BNGetFriendInfoByID(_G.GetAutoCompletePresenceID(tarAttribute))
			if not battletag then 
				battletag = plr 
			end
			tarAttribute = battletag
		end
		
		if dbg.chatHistory[tarAttribute] and dbg.chatHistory[tarAttribute].message[1] then
			if tarAttribute == OEBName and dbg == OEBNamePath then return end  -- stops mod:ChatEdit_Update's rapid firing
			OEBName, OEBNamePath = tarAttribute, dbg
			mod:ShowFrame()
		else 
			mod:ChatEdit_OnHide() 
		end
	else 
		mod:ChatEdit_OnHide() 
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:ChatEdit_OnHide()
	if self.editboxTitle then
		self.editboxTitle:SetScript("OnLeave", nil)
		self.editboxTitle:SetScript("OnUpdate", nil)
		self.editboxTitle:Hide()
		QTC:Release(self.editboxTitle)
		self.editboxTitle = nil
	end
	editboxFrame:Hide()
	OEBName = nil
	OEBNamePath = nil
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:OnSkinUpdate()
	SoDWhisper:SkinMyFrame(editboxFrame)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:CreateFrame()
	local frame = SoDWhisper:CreateMyFrame("EDITBOX", mod.db.profile.frameWidth)
	
	frame:EnableMouse(true)
	frame:SetResizable(true)
	frame.dragRight = mod:AddDragFrame(frame, "BOTTOMRIGHT", "BOTTOMRIGHT", "TOPRIGHT", "TOPRIGHT", "RIGHT")
	frame.dragLeft  = mod:AddDragFrame(frame, "BOTTOMLEFT", "BOTTOMLEFT", "TOPLEFT", "TOPLEFT", "LEFT")
	frame:SetMovable(true)
	frame:SetScript("OnMouseDown", function(self, arg1) 
		if not mod.db.profile.locked and arg1 == "LeftButton" then
			self:StartMoving()
			self.isMoving = true
		end
	end)
	frame:SetScript("OnMouseUp", function(self, arg1) 
		if not mod.db.profile.locked and arg1 == "LeftButton" then
			self:StopMovingOrSizing()
			self.isMoving = false
			mod:SavePosition()
		end
	end)
	return frame
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:SetupEditboxtTitle()	
	local dbH = OEBNamePath.chatHistory[OEBName]
	local tempIsBT = (OEBNamePath == dbg and true) or nil
	
	local editboxTitle = QTC:Acquire("SoDWhisper_EditboxTitle", 6, "LEFT", "RIGHT", "RIGHT", "RIGHT", "RIGHT", "RIGHT")
	mod.editboxTitle = editboxTitle
	editboxTitle:SetFrameStrata("HIGH")
	editboxTitle:Clear()
	editboxTitle:SetScale(mod.db.profile.titlescale)
	local bg = SoDWhisper.media:Fetch('background', SoDWhisper.db.profile.paneBackground, "Blizzard Tooltip")
	local ed = SoDWhisper.media:Fetch('border', SoDWhisper.db.profile.paneBorder, "Blizzard Tooltip")
	editboxTitle:SetBackdrop({
		bgFile = bg, tile = true, tileSize = 16,
		edgeFile = ed, edgeSize = 16,
		insets = {left = 3, right = 3, top = 3, bottom = 3},
	})
	local c,d = SoDWhisper.db.profile.paneColor, SoDWhisper.db.profile.paneBorderColor
	editboxTitle:SetBackdropColor(c[1], c[2], c[3], c[4])
	editboxTitle:SetBackdropBorderColor(d[1], d[2], d[3], d[4])
	
	editboxTitle:AddHeader()
	
	editboxTitle:SetCell(1, 1, copyIcon, nil, "RIGHT")
	editboxTitle:SetCellScript(1, 1, "OnMouseDown", mod.HandleTTClick, "copyIcon")
	
	if tempIsBT then
		local templvl, tempToonDN = "", ""
		if dbH.online then
			if dbH.client == BNET_CLIENT_WOW then
				templvl = SoDWhisper:GetColoredLevel(dbH.level, dbH.online, true, true)
			end	
			tempToonDN = " - "..SoDWhisper:GetDisplayNameForTooltip(OEBName, OEBNamePath, true, true, true, 16)
		end
		editboxTitle:SetCell(1, 2, templvl..SoDWhisper:GetDisplayNameForTooltip(OEBName, OEBNamePath, true, nil, true, 16)..tempToonDN, nil, "LEFT")
	else 
		editboxTitle:SetCell(1, 2, SoDWhisper:GetColoredLevel(dbH.level, dbH.online, nil, true)..SoDWhisper:GetDisplayNameForTooltip(OEBName, OEBNamePath, nil, nil, nil, 16), nil, "LEFT")
	end
	editboxTitle:SetCellScript(1, 2, "OnMouseDown", function()
		SoDWhisper:UpdateChangePlayerName(SoDWhisper.createDisplayName, OEBName, OEBNamePath)
	end)
	
	editboxTitle:SetCell(1, 3, inviteIcon, nil, "RIGHT", tempIsBT and 3 or nil)
	editboxTitle:SetCellScript(1, 3, "OnMouseDown", mod.HandleTTClick, "inviteIcon")
	
	if not tempIsBT then
		editboxTitle:SetCell(1, 4, addfriendIcon, nil, "RIGHT")
		editboxTitle:SetCellScript(1, 4, "OnMouseDown", mod.HandleTTClick, "addfriendIcon")
		
		editboxTitle:SetCell(1, 5, ignoreIcon, nil, "RIGHT")
		editboxTitle:SetCellScript(1, 5, "OnMouseDown", mod.HandleTTClick, "ignoreIcon")
	end
	
	editboxTitle:SetCell(1, 6, locationIcon, nil, "RIGHT")
	editboxTitle:SetCellScript(1, 6, "OnMouseDown", mod.HandleTTClick, "locationIcon")
	editboxTitle:SetCellScript(1, 6, "OnEnter", function()
					local titletooltip = QTC:Acquire("SoDWhisper_TitleTooltip", 1, "CENTER")
					mod.titletooltip = titletooltip
					titletooltip:Clear()
					titletooltip:SetScale(mod.db.profile.titlescale)
					titletooltip:ClearAllPoints()
					titletooltip:SetPoint(mod.db.profile.growUp and "BOTTOMRIGHT" or "TOPRIGHT", editboxTitle, mod.db.profile.growUp and "TOPRIGHT" or "BOTTOMRIGHT")
					
					local bg = SoDWhisper.media:Fetch('background', SoDWhisper.db.profile.paneBackground, "Blizzard Tooltip")
					local ed = SoDWhisper.media:Fetch('border', SoDWhisper.db.profile.paneBorder, "Blizzard Tooltip")
					titletooltip:SetBackdrop({
						bgFile = bg, tile = true, tileSize = 16,
						edgeFile = ed, edgeSize = 16,
						insets = {left = 3, right = 3, top = 3, bottom = 3},
					})
					local c,d = SoDWhisper.db.profile.paneColor, SoDWhisper.db.profile.paneBorderColor
					titletooltip:SetBackdropColor(c[1], c[2], c[3], c[4])
					titletooltip:SetBackdropBorderColor(d[1], d[2], d[3], d[4])
					
					if dbH.zone and dbH.zone ~= "" then
						titletooltip:AddHeader(SoDWhisper:GetZoneColor(dbH.zone))
					end
					if dbH.guild and dbH.guild ~= "" then
						titletooltip:AddHeader("<"..dbH.guild..">")
					end
					if tempIsBT and dbH.online and dbH.client == BNET_CLIENT_WOW then
						titletooltip:AddHeader(SoDWhisper:GetRealmForTooltip(OEBName, false, 16, OEBNamePath))
					end
					
					if titletooltip:GetLineCount() ~= 0 then
						titletooltip:Show()
					end
					titletooltip:SetClampedToScreen(true)
		end)
	editboxTitle:SetCellScript(1, 6, "OnLeave", function()
				mod.titletooltip:Hide()
				QTC:Release(mod.titletooltip)
				mod.titletooltip = nil
	end)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:HandleTTClick(name, button)
	if not name then return end
	
	if name == "inviteIcon" then
		if OEBNamePath.chatHistory[OEBName].pID then
			_G.FriendsFrame_BattlenetInvite(nil, OEBNamePath.chatHistory[OEBName].pID)
		else
			_G.InviteUnit(OEBName)
		end
	elseif name == "copyIcon" then
		SoDWhisper:ShowLogFrame(SoDWhisper.LogFrame, OEBName, OEBNamePath)
	elseif name == "addfriendIcon" then
		if _G.BNFeaturesEnabledAndConnected() then
			_G.AddFriendEntryFrame_Collapse(true);
			_G.AddFriendFrame.editFocus = _G.AddFriendNameEditBox;
			_G.StaticPopupSpecial_Show(_G.AddFriendFrame);
			if _G.GetCVarBool("addFriendInfoShown") then
				_G.AddFriendFrame_ShowEntry();
			else
				_G.AddFriendFrame_ShowInfo();
			end
		else
			_G.StaticPopup_Show("ADD_FRIEND");
		end	
	elseif name == "ignoreIcon" then
		_G.AddOrDelIgnore(OEBName)
	elseif name == "locationIcon" then
		if OEBNamePath == dbr then
			local wholib = SoDWhisper
			wholib:Who(OEBName, {queue = wholib.WHOLIB_QUEUE_USER})
		end
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:UpdateOpenEditBox()
	if SoDWhisper.lastSender == OEBName and SoDWhisper.lastSenderPath == OEBNamePath then
		SoDWhisper:UpdateMyFrame(editboxFrame, SoDWhisper.lastSender, SoDWhisper.lastSenderPath, mod.db.profile.paneSortDown)
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:ShowFrame()
	if not OEBName or not OEBNamePath or not OEBNamePath.chatHistory[OEBName].message[1] then return end
	
	mod:SetupEditboxtTitle() 	-- Must be before SetPostion	
	mod:SetPosition() 			-- Must be before UpdateMyFrame function
	SoDWhisper:UpdateMyFrame(editboxFrame, OEBName, OEBNamePath, mod.db.profile.paneSortDown)	
	
	editboxFrame:Show()
	mod.editboxTitle:Show()
	mod.editboxTitle:SetClampedToScreen(true)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:ResetPosition()
	mod.db.profile.frameWidth = 400
	mod.db.profile.frameLeft = 10
	mod.db.profile.frameTop = 370
	mod.db.profile.frameBottom = 400
	mod:SetPosition()
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:SetPosition()
	editboxFrame:ClearAllPoints()
	local ChatFrameEditBox = _G.ChatEdit_GetActiveWindow()
	if self.db.profile.snapToEditbox and ChatFrameEditBox then
		mod.db.profile.frameWidth = ChatFrameEditBox:GetRight() - ChatFrameEditBox:GetLeft()
		if self.db.profile.growUp then
			editboxFrame:SetPoint("BOTTOMLEFT",ChatFrameEditBox,"TOPLEFT", 0, 0)
		else
			editboxFrame:SetPoint("TOPLEFT",ChatFrameEditBox,"BOTTOMLEFT", 0, 0)
		end
	else
		if self.db.profile.growUp then
			editboxFrame:SetPoint("BOTTOMLEFT",UIParent,"BOTTOMLEFT", mod.db.profile.frameLeft, mod.db.profile.frameBottom)
		else
			editboxFrame:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT", mod.db.profile.frameLeft, mod.db.profile.frameTop)
		end
	end
	editboxFrame:SetWidth(mod.db.profile.frameWidth)
	
	if mod.editboxTitle then
		mod.editboxTitle:ClearAllPoints()
		mod.editboxTitle:SetPoint(mod.db.profile.growUp and "BOTTOMRIGHT" or "TOPRIGHT", editboxFrame, mod.db.profile.growUp and "TOPRIGHT" or "BOTTOMRIGHT")
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
-- from Whisp
function mod:SavePosition()
	mod.db.profile.frameWidth = editboxFrame:GetRight() - editboxFrame:GetLeft()
	mod.db.profile.frameLeft = editboxFrame:GetLeft()
	mod.db.profile.frameTop = editboxFrame:GetTop()
	mod.db.profile.frameBottom = editboxFrame:GetBottom()
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:AddDragFrame(parent, anchor1, align1, anchor2, align2, direction)
	local frame = CreateFrame("Frame", nil, parent)
	frame:Show()
	frame:SetFrameLevel(parent:GetFrameLevel() + 10)
	frame:SetWidth(16)
	frame:SetPoint(anchor1, parent, align1, 0, 0)
	frame:SetPoint(anchor2, parent, align2, 0, 0)
	frame:EnableMouse(true)
	frame:SetScript("OnMouseDown", function(self, arg1) 
		if not mod.db.profile.locked and arg1 == "LeftButton" then
			self:GetParent().isResizing = true
			self:GetParent():StartSizing(direction)
		end 
	end)
	frame:SetScript("OnMouseUp", function(self, arg1) 
		if self:GetParent().isResizing == true then 
			self:GetParent():StopMovingOrSizing()
			self:GetParent().isResizing = false
			mod:SavePosition()
			SoDWhisper:SendMessage("SoDWhisper_MESSAGE")
		end
	end)
	return frame
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:GetOptions()
	return options
end
------------------------------------------------------------------------------------------------------------------------------------------------------------
function mod:Info()
	return L["This plugin will show your current conversation when you are sending a whisper to someone."]
end
-- end Whisp